// MET2mml.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#pragma warning(disable : 4996)

// #define STDOUT

#define MAX_PATH 256

#define	TIME_BASE 48
#define	BARS_PER_LINE 8
#define MML_ACCENT_ON "k+20"
#define MML_ACCENT_OFF "k-20"
//#define GATE_NORMAL 6
#define GATE_NORMAL 8 // ##210901
#define STD_VOLUME 4  // ##210901
#define LEN_DEFAULT 16
#define	TEMPO_DEFAULT 120

#pragma pack(push, 1)

struct MetMusicHeader
{
	short dwSize;
	short nPartCount;
	short nBarCount;
	char  nBarLines;
	char  nBarSkip;
	char  szTitle[24];
};

#pragma pack(pop)

//static const int nVolumeTable[8] = { 16, 32, 48, 64, 80, 96, 112, 128 };  ##210901
static const int nVolumeTable[8] = { 32, 48, 64, 76, 88, 100, 112, 127 };
static int nGateTable[4] = { GATE_NORMAL, 8, 4, 2} ;
//static const char *szKeyTable[12] = { "c", "c+", "d", "d+", "e", "f", "f+", "g", "g+", "a", "a+", "b" };
static const char *szKeyTables[2][12] = {
	{ "c", "d-", "d", "e-", "e", "f", "f+", "g", "a-", "a", "b-", "b" },		// Major
	{ "c", "c+", "d", "d+", "e", "f", "f+", "g", "g+", "a", "b-", "b" } };	// Minor


static const char szMetExt[] = ".msc";
static const char szVoiceExt[] = ".vce";
static const char szMMLExt[] = ".mml";
static const char szVoiceTableExt[] = ".vct"; // ##140520

static const char *szOctaveSymbols[5] = { "-15ma", "-8va", "oct0", "+8va", "+15ma" }; // ##140519 fix:Octave

FILE* openfile(const char *filename, const char *mode)
{
	FILE *f = fopen(filename, mode);
	if(!f)
		printf("cannot open '%s' for <%s>\n", filename, mode);
	return f;
}

long file_length(FILE *f)
{
	long len = 0;
	
	fseek(f, 0, SEEK_END);
	len = ftell(f);
	fseek(f, 0, SEEK_SET);
	return len;
}

BOOL Met2MML(MetMusicHeader* pMetHeader, FILE *f);
BOOL Met2MMLTrack(MetMusicHeader* pMetHeader, BYTE* pMetPart, int nPart, FILE *f);
BYTE* pMetData = NULL;
BYTE* pVoiceData = NULL;
int nVceCount;
int nTempo;

// ##140520
#define VCE_COUNT 128
#define MAX_BUFFER 256
static const char szSep[] = ",\t\r\n";
static const char szVctSep = '\t';

BYTE bVceUse[VCE_COUNT];

// char *pVceTable[VCE_COUNT]; ##210901

struct VceTable {
	char* prog;
	short volume;
	short octave;
};
VceTable vceTable[VCE_COUNT];

short ReadValue(char *p);
void SetKeyTable(int note, bool major);
BYTE FixNatural(BYTE b1);
int RangeVolume(BYTE b, short vceVolume);
BYTE GetNextNote(BYTE* pMetPart, BYTE* pMetPartEnd, BOOL is3ch);
BYTE CheckNoteModifier(BYTE b1);
void ClearNoteModifier();

char szVctPathName[MAX_PATH];

//========================================================================
// MET2mml main

int main(int argc, char* argv[])
{
	if(argc <  2) {
		printf("MET music file to MML text converter Ver. 0.10 by c.mos\r\n" \
			   "  usage: MET2mml <MET music file(AAA.msc)> [MML files path]\r\n" \
			   "         (read AAA.vce and read/write AAA.vct file)\r\n");
		return EXIT_FAILURE;
	}

	char szPathName[MAX_PATH];

	strcpy(szPathName, argv[1]);
	char *ext =  strrchr(szPathName, '.');
	if(!ext) strcat(szPathName, szMetExt);

	// .msc read
	FILE *f = openfile(szPathName, "rb");
	if(!f) return EXIT_FAILURE;
	long nFileLen = file_length(f);
	pMetData = new BYTE[nFileLen];
	fread(pMetData, 1, nFileLen, f);
	fclose(f);

	// .vce read
	ext =  strrchr(szPathName, '.');
	if(ext) *ext = '\0';
	strcat(szPathName, szVoiceExt);
	f = openfile(szPathName, "rb");
	if(!f) return EXIT_FAILURE;
	nFileLen = file_length(f);
	pVoiceData = new BYTE[nFileLen];
	fread(pVoiceData, 1, nFileLen, f);
	fclose(f);
	nVceCount = *(short*)(pVoiceData + 6);

	// .vct read ##140520
	memset(bVceUse, FALSE, VCE_COUNT);
//	memset(pVceTable, NULL, sizeof(char*) * VCE_COUNT);
	memset(vceTable, 0, sizeof(vceTable));

	ext =  strrchr(szPathName, '.');
	*ext = '\0';
	strcat(szPathName, szVoiceTableExt);
	strcpy(szVctPathName, szPathName);

	BOOL bReadVct = FALSE;
	f = openfile(szPathName, "r");
	if (f) {
		bReadVct = TRUE;
		char buffer[MAX_BUFFER];
		while (fgets(buffer, MAX_BUFFER, f)) {
			char *p = strtok(buffer, szSep);
			int n = atoi(p);
			p = strtok(NULL, szSep);
			char *q = new char[strlen(p) + 1];
			strcpy(q, p);
			vceTable[n - 1].prog = q; // ##210901
			p = strtok(NULL, szSep);
			while (p) {
				if (*p == 'V' || *p == 'v') {
					vceTable[n - 1].volume = ReadValue(p + 1);
				} else if (*p == 'O' || *p == 'o') {
					vceTable[n - 1].octave = ReadValue(p + 1);
				}
				p = strtok(NULL, szSep);
			}
		}
		fclose(f);
	}

	// MMLo̓tH_
	if(argc > 2)
		strcpy(szPathName, argv[2]);
	else {
		char *ext =  strrchr(szPathName, '.');
		if(ext) *ext = '\0';
	}
	_mkdir(szPathName);
	char* pszMMLFileName = szPathName + strlen(szPathName);
	*pszMMLFileName++ = '\\';
	*pszMMLFileName = '\0';

	if(strcmp((const char*)pMetData, "MSC")) {
		printf("Illegual MET music file header!\r\n");
	} else {
		int nMusicCount = *(short*)(pMetData + 6);
		BYTE* pHeader = pMetData + 16;
		for_to(i, nMusicCount) {
//			strncpy(pszMMLFileName, ((MetMusicHeader*)pHeader)->szTitle, 24);
//			strcat(szPathName, szMMLExt);
			sprintf(pszMMLFileName, "%02d %s%s", i + 1, ((MetMusicHeader*)pHeader)->szTitle, szMMLExt); // ##210901
#ifdef STDOUT
			FILE *f = stdout;
#else
			FILE *f = openfile(szPathName, "w");
#endif
			if(!f) return EXIT_FAILURE;
			nTempo = 0;
			if(!Met2MML((MetMusicHeader*)pHeader, f)) break;
#ifndef STDOUT
			fclose(f);
#endif
			pHeader += 	((MetMusicHeader*)pHeader)->dwSize;
		}
	}

	// .vct write ##140520
	if (!bReadVct) {
		FILE *f = openfile(szVctPathName, "w");
		if(f) {
			for_to(i, nVceCount) {
				if (bVceUse[i]) {
					char *pszVoice = (char *)((pVoiceData + 16) + i * 64);
					fprintf(f, "%d%c@0%c%s\n", i + 1, szVctSep, szVctSep, pszVoice);
				}
			}
			fclose(f);
		}
	}

	delete pMetData;
	delete pVoiceData;
	for_to(i, VCE_COUNT) {
		delete vceTable[i].prog;
	}

	return EXIT_SUCCESS;
}

short ReadValue(char* p) // ##210901
{
	short value = 0;
	char c;
	while (c = *p++) {
		if (c == '+') {
			value++;
		} else if (c == '-') {
			value--;
		} else if (c >= '1' && c <= '9') {
			if (value >= 0) {
				value = c - '0';
			} else {
				value = -(c - '0');
			}
		}
	}
	return value;
}

//========================================================================
// Convert Met music data to MML text

int s_count = 0; // ##140519 D.S.Patch

BOOL ConfirmRepeat(BYTE b1, BYTE* pMetPart, BYTE* pMetPartEnd, BOOL bChord) // ##140519 fix:Repeat
{
	if (b1 != 0x40 && b1 != 0x80) return TRUE;
	while (pMetPart < pMetPartEnd) {
		BYTE c0 = *pMetPart++;
		BYTE c1 = *pMetPart++;
		if (c0 == 0x04) {
			if (c1 != 0x09) return FALSE; // ##140526
		} else if (c0 == 0x0E) {
			if (c1 == b1) return FALSE;
		} else if (c0 == 0x0F) {
			c1 &= 0xE0; // ##140521
			if (c1 == b1) return TRUE;
		}
		if (c0 > 0x0F && bChord) {
			pMetPart += 2;
		}
	}
	return FALSE;
}

BOOL Met2MML(MetMusicHeader* pMetHeader, FILE *f)
{
	fprintf(f, "#title \"%s\"\n", pMetHeader->szTitle);
	fprintf(f, "#timebase %d\n", TIME_BASE);
//	printf("title: %s\n", pMetHeader->szTitle);

	BYTE* pMetPart = (BYTE*)(pMetHeader + 1);
	for_to(i, pMetHeader->nPartCount) {
		s_count = 0; // ##140519 D.S.Patch
//		printf("  track[%d] offset: %04X\n", i + 1, pMetPart - pMetData);
		if(!Met2MMLTrack(pMetHeader, pMetPart, i, f)) return FALSE;
		pMetPart += *((short*)pMetPart);
	}
	return TRUE;
}

BOOL Met2MMLTrack(MetMusicHeader* pMetHeader, BYTE* pMetPart, int nPart, FILE *f)
{
//	if(strcmp(pMetHeader->szTitle, "Cordarn") == 0 && nPart == 3) {
//		TRACE("!");		
//	}
	BYTE mode = *(pMetPart + 2);
	int  chord = (mode>>2)&3;
	BOOL noise = (mode & 3) == 3;
	fprintf(f, "\n;===== %s part%d%s =====\n",  (noise ? "NOISE" : (mode & 0x80) ? "PSG" : "FM"), nPart+1, chord ? " (3ch)" : ""); // ##140518
	fprintf(f, "%c C%d v%d l16 Q%d ", 'A'+ nPart, nPart + 1, nVolumeTable[STD_VOLUME], GATE_NORMAL);
	if (mode & 0x80) {
		fputs("@80 ", f); // ##210901
	}

	BOOL bLF = TRUE;
	BOOL bBar = FALSE;
	int nBarLength = TIME_BASE * 4;	// 4/4
	int nBarCount = 0; // ##210901
	int nTimeInBar = 0;
	int nTimeInRepeat = 0;
	int bAccent = 0;
	// int nLen = 4;
	int nVolume = STD_VOLUME;
	int nGate = 0;
	int nOctave = -1;
	int nOctaveBase = 0;
	int nRepeatNest = 0;
	BOOL bTai = FALSE;
	BOOL bTaiNext = FALSE;
	nGateTable[0] = GATE_NORMAL;
	BYTE* pMetPartEnd = pMetPart + *(short*)pMetPart;
	pMetPart += 4;

	VceTable* pVceTable = &vceTable[0]; // ##210901

	SetKeyTable(0, TRUE);
	ClearNoteModifier();

	while(pMetPart < pMetPartEnd) {
		BYTE b0 = *pMetPart++;
		BYTE b1 = *pMetPart++;
		BYTE b1Next = GetNextNote(pMetPart, pMetPartEnd, chord > 0); // ##210901

		if(bLF) {
			if (b0 != 4 && b0 != 15) { // Repeat
				fprintf(f, "\n%c ", 'A'+ nPart);
			}
			bLF = FALSE;
			nBarCount = 0;
			bBar = FALSE;
		} else if (bBar) {
			if (b0 != 4 && b0 != 15) {
				fputs(" | ", f);
			}
			bBar = FALSE;
		}

		if(bAccent == -1) {				// ## 140518 ACCENT
			if (b0 != 11) {
				fprintf(f, MML_ACCENT_OFF);
				bAccent = FALSE;
			}
		}

		if(b0 <= 0x0F) { // control data
			if(b0 == 0) break;
			switch(b0) {
			case 2: {// Key sig.
				BOOL isMajor = (b1 & 0x80) == 0;
				char c = isMajor ? 'J' : 'I';
				int note = b1 & 15;
				fprintf(f, "K%c%s ", c, szKeyTables[isMajor ? 0 : 1][note]);
				SetKeyTable(note, isMajor);
				break;
			}
			case 3: {// Beats
				int nBeats = (b1 >> 3) + 1;
				int nUnit = 1 << ((b1 & 7) - 1);
				if (b1 == 0) {
					nBeats = nUnit = 4; // ##210901
				}
				fprintf(f, "/* %d/%d */ ", nBeats, nUnit);
				nBarLength = ((TIME_BASE * 4) / nUnit) * nBeats;
				break;
			}
			case 4: // Repeat Controls
				switch(b1 & 0x0F) {
				case 2: {
					if(--nRepeatNest < 0)
						nRepeatNest = 0;
					else {
						if (nGate != 0 || nGateTable[0] != GATE_NORMAL) { // ##210901
							nGate = 0;
							nGateTable[0] = GATE_NORMAL;
							fprintf(f, " Q%d", GATE_NORMAL);
						}
						fputc(']', f);
						int nCount = b1 >> 4;
						if(nCount)
							fprintf(f, "%d", nCount + 2);
						bLF = TRUE;
						nTimeInBar = 0;
						ClearNoteModifier();
						nOctave = -1;
					}
					break;
				}
				case 9:
					fputc(':', f);
					break;
				case 12:
					fputc('[', f);
					nOctave = -1;
					nRepeatNest++;
					break;
				case 5: { // D.S. ##140519 D.S.Patch
					if (s_count == 0) break;
					s_count--;
					fputc(']', f);
					bLF = TRUE;
					nTimeInBar = 0;
					ClearNoteModifier();
					nOctave = -1;
					break;
				}
				case 8: // S ##140519 D.S.Patch
					if (s_count != 0) break;
					s_count++;
					fputc('[', f);
					nOctave = -1;
					break;
				}
				break;
			case 5: // Tempo
				fprintf(f, "t%d ", b1);
				if(!nTempo) nTempo = b1;
				break;
			case 6: // Octave
				// fprintf(f, "/* oct%d */ ", (char)b1); // ##140519 fix:Octave
				nOctaveBase = (char)b1;
				fprintf(f, "/* %s */", szOctaveSymbols[nOctaveBase + 2]);
				break;
			case 7: {// Voice
				char *pszVoice = (char *)((pVoiceData + 16) + b1 * 64);
				char vcename[32];
				if (b1 >= nVceCount) {
					sprintf(vcename, "#%d", b1); // ##210901
					pszVoice = vcename;
				}
				short vceVolumeOld = pVceTable->volume;
				pVceTable = &vceTable[b1];
				fprintf(f, "%s/* %s */ ", pVceTable->prog ? pVceTable->prog : "@0", pszVoice);  // ##140520
				bVceUse[b1] = TRUE;
				if (pVceTable->volume != vceVolumeOld) {
					int index = RangeVolume(nVolume, pVceTable->volume);
					fprintf(f, "v%d ", nVolumeTable[index]);
				}
				break;
			}
			case 8: // Gate/Step
				fprintf(f, "Q%d", b1 + 1);
				nGateTable[0] = b1 + 1;
				break;
			case 9: // LFO
				fprintf(f, "/* LFO %s */ ", b1 ? "ON" : "OFF");
				break;
			case 10: // Volume
			{
				nVolume = b1;
				int index = RangeVolume(b1, pVceTable->volume);
				fprintf(f, "v%d ", nVolumeTable[index]);
				break;
			}
			case 11: // Accent
				if(!bAccent)
					fprintf(f, MML_ACCENT_ON);
				bAccent = TRUE;
				break;
			case 12: // Port
				fprintf(f, "/* Port */ ");
				break;
			case 14: // Repeat Begin
				if (ConfirmRepeat(b1, pMetPart, pMetPartEnd, chord > 0)) { // ##140519 fix:Repeat
					fprintf(f, "[");
					nOctave = -1;
				} else {
					printf("*** skip '[' in [%s] part#%d\n", pMetHeader->szTitle, nPart+1);
				}
				nTimeInRepeat = 0;
				break;
			case 15: // Repeat End
				int count = (b1 & 0x3F) + 1;
				fprintf(f, "]%d", count);
				nTimeInBar = nTimeInRepeat * count;
				nTimeInRepeat = 0;
				break;
			}
		} else {
			if(!nTempo) {
				fprintf(f, "t%d", TEMPO_DEFAULT);
				nTempo = TEMPO_DEFAULT;
			}
//			if(nTimeInBar == 0)
//				fprintf(f, "\n%c ", 'A'+ nPart);
/*			if(bAccent == -1) {				// ## 140518 ACCENT
				fprintf(f, MML_ACCENT_OFF);
				bAccent = FALSE;
			}*/
			else if(bAccent == TRUE)
				bAccent = -1;
			if (b1 != 0xFF) { // ##140518
				int nOctaveNew = (b1 >> 3) & 7;
				nOctaveNew += nOctaveBase; // ## 140519 fix:Octave
				nOctaveNew += pVceTable->octave; // ##210901
				if(nOctave != nOctaveNew) {
					if(nOctave != -1 && nOctave - nOctaveNew == 1)
						fputc('<', f);
					else if(nOctave != -1 && nOctave - nOctaveNew == -1)
						fputc('>', f);
					else
						fprintf(f, "o%d", nOctaveNew);
					nOctave = nOctaveNew;
				}
			}
			int nGateNew = (b0 >> 5) & 3;
			if(nGateNew == 1 && b1  == b1Next) {	// ^C ##210901
				bTaiNext = TRUE;
			} else if(!bTai) {
				if(nGate != nGateNew) {
					nGate = nGateNew;
					fprintf(f, "Q%d", nGateTable[nGate]);
				}
			}
			int nNote = (chord == 0) ? 1 : 3; // ##140518 fix:PSG123
			BYTE* pMetPartNext = (chord == 0) ? pMetPart : pMetPart + 2;
			for(int i = 0; i < nNote; i++, b1 = *pMetPart++) {
				if(b1 == 0xFF || bTai) {
					if (i == 0) {
						fputc('r', f);
					}
					break;
				} else {
					char cTone = 'c' + (b1 & 7);
					if(cTone > 'g') cTone -= 'h' - 'a';
					fputc(cTone, f);
					BYTE b = FixNatural(b1);
					if (b != b1) {
						printf("*** fix natural in [%s] part#%d\n", pMetHeader->szTitle, nPart + 1);
					}
					b = CheckNoteModifier(b);
					switch(b & 0xC0) {
						case 0x40: fputc('+', f); break;
						case 0x80: fputc('-', f); break;
						case 0xC0: fputc('*', f); break;
						break;
					}
					if (i < nNote - 1 && *pMetPart != 0xFF) { // ##140518 fix:PSG123
						fputc('0', f);
					}
				}
			}
			pMetPart = pMetPartNext;
			bTai = FALSE;
			int nLen = b0 & 7;
			int nTime;
			if(nLen == 0) {
				fprintf(f, "%%%d", nBarLength);
				nTime = nBarLength;
			} else {
				nLen = 1 << (nLen - 1);
				nTime = TIME_BASE * 4 / nLen;
				int mod = (b0 >> 3) & 3;
				if(mod != 3) {
					if(nLen != LEN_DEFAULT)
						fprintf(f, "%d", nLen);
					if(mod == 1) {
						fputc('.', f);
						nTime = nTime + nTime / 2;
					} else if(mod == 2) {
						fputs("..", f);
						nTime = nTime + nTime / 2 + nTime / 4;
					}
				} else { // OA
					nTime = nTime * 2 / 3;
					fprintf(f, "%%%d", nTime); // // ##140519 fix
				}
			}
			if(bTaiNext) {
				fputc('&', f);
				bTaiNext = FALSE;
				bTai = TRUE;
			}
			nTimeInBar += nTime;
			nTimeInRepeat += nTime;
			if(nTimeInBar >= nBarLength) {
				if (++nBarCount >= BARS_PER_LINE && !bTai) {
					nBarCount = 0;
					bLF = TRUE;
				}
				nTimeInBar = 0;
				ClearNoteModifier();
				bBar = TRUE;
			}
		}
	}
	fputc('\n', f);
	return TRUE;
}

// ##210901

BYTE GetNextNote(BYTE* pMetPart, BYTE* pMetPartEnd, BOOL is3ch)
{
	if (is3ch) pMetPart += 2;
	while (pMetPart < pMetPartEnd) {
		BYTE b0 = *pMetPart++;
		BYTE b1 = *pMetPart++;
		if (b0 > 0x0F) return b1;
	}
	return 0xFF;
}

#define NOTE_COUNT 7
static BYTE noteTable[NOTE_COUNT];

BYTE CheckNoteModifier(BYTE b1)
{
	int note = b1 & 7;
	if (b1 & 0xC0) {
		noteTable[note] = b1;
	} else if (noteTable[note] & 0xC0) {
		b1 = noteTable[note];
	}
	return b1;
}

void ClearNoteModifier()
{
	memset(noteTable, 0, NOTE_COUNT);
}

int RangeVolume(BYTE b, short vceVolume)
{
	int index = b + vceVolume;
	if (index < 0) {
		index = 0;
	}
	else if (index > 7) {
		index = 7;
	}
	return index;
}

//========================================================================
//* fix NATURAL by MMLParser.cpp ##210901

#define OCTAVE_STEPS 12

char keyTable[OCTAVE_STEPS];

#define MAX_KEY_TABLE 6

const char sharpIndices[MAX_KEY_TABLE] = { 5, 0, 7, 2, 9, 4 };
const char flatIndices[MAX_KEY_TABLE] = { 11, 4, 9, 2, 7, 0 };

void SetKeyTable(int note, bool major) {
	for (int i = 0; i < OCTAVE_STEPS; i++) {
		keyTable[i] = 0;
	}
	int n = major ? 0 : 9;
	int s = -1; // circle of 5th (clockwise)
	for (; s < MAX_KEY_TABLE; s++) {
		if (n == note) break;
		if ((n += 7) >= OCTAVE_STEPS) {
			n -= OCTAVE_STEPS;
		}
	}
	if (s == -1) return;
	n = major ? 0 : 9;
	int f = -1; // circle of 5th (anticlockwise)
	for (; f < MAX_KEY_TABLE; f++) {
		if (n == note) break;
		if ((n -= 7) < 0) {
			n += OCTAVE_STEPS;
		}
	}
	if (s <= f) {
		for (int i = 0; i <= s; i++) {
			keyTable[sharpIndices[i]] = 1;
		}
	}
	else {
		for (int i = 0; i <= f; i++) {
			keyTable[flatIndices[i]] = -1;
		}
	}
}

const char noteToIndex[] = { 0, 2, 4, 5, 7, 9, 11 };

BYTE FixNatural(BYTE b1)
{
	int n = keyTable[noteToIndex[b1 & 7]];
	BYTE f = b1 & 0xC0;
	if (n > 0 && f == 0x80 || n < 0 && f == 0x40) {
		b1 = 0xC0 | (b1 & ~0xC0);
	}
	return b1;
}
